/***************************************************************************
  
  gb.gl.h
  
  (c) 2005-2007 Laurent Carlier <lordheavy@users.sourceforge.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
  
***************************************************************************/

#ifndef __GB_GL_H
#define __GB_GL_H

#include "gambas.h"

#define GL_INTERFACE_VERSION 1

typedef
  struct {
    intptr_t version;
    // Must be called after the context is init !
    //** Perhaps also when context is changed but not tested **
    bool (*Init)(void);

    void *_null;
  }
  GL_INTERFACE;

#endif
