/***************************************************************************
  
  ggambastag.h
  
  (c) Benoît Minisini <benoit.minisini@gambas-basic.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
  
***************************************************************************/

#ifndef __GGAMBASTAG_H
#define __GGAMBASTAG_H

#include "gtag.h"

class gGambasTag: public gTag
{
public:
	gGambasTag(void *p) : gTag(p) { }
	virtual ~gGambasTag() {}
	virtual void ref(void *v) { GB.Ref(v); }
	virtual void unref(void *v) { GB.Unref((void **)&v); }
};

#endif
