/***************************************************************************
  
  deletemap.h
  
  (c) Benoît Minisini <benoit.minisini@gambas-basic.org>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
  
***************************************************************************/

#ifndef __DELETEMAP_H
#define __DELETEMAP_H

typedef
  void *DELETE_MAP;
  
int DELETE_MAP_virtual_to_real(DELETE_MAP *dmap, int vpos);
int DELETE_MAP_real_to_virtual(DELETE_MAP *dmap, int rpos);
void DELETE_MAP_add(DELETE_MAP **dmap, int vpos);
void DELETE_MAP_free(DELETE_MAP **dmap);

#endif
